;	COMP	Replacement for the messy-dos command of the same name.

;	Usage:	comp [/#] file1 [file2]
;		where file1 and file2 may be directories or may contain
;		wildcards (but not both).  The default for file2 is the current
;		directory on the current disk.  The switch gives the maximum
;		number of errors to report per file.  Default is 65535 if file1
;		refers to a single file, 10 otherwise.
;		The '/' in the switch refers to the switch character.

;	Author:  Paul Vojta
;	Translated from A86 to NASM assembly language dialect by Eric Auer
;	(because Steve Nickolas wanted to tune COMP but has no A86) 8/2003

; %define cloneA86 1	; make .com files almost equal, apart from
			; 18 cases where NASM and A86 encode the
			; same operation in different same length
			; opcodes. Makes the .com 9 bytes longer.
	; to verify the almost-equal content, do:
	; ndisasm one.com > a
	; ndisasm other.com > b
	; wdiff -l a b | less

stdout	equ	1
stderr	equ	2

CR	equ	13
LF	equ	10
TAB	equ	9

	; the next are inside the "default FCB" of DOS PSP, see int 21.26
	; and RBIL 61 table 01345, but are not used in FCB related ways!
	; Paul just saves some RAM by putting variables into PSP.

fn1	equ	5ch	; offset of pointer to filename portion of string 1
str2	equ	5eh	; offset of pointer to (?) string 2
fn2	equ	60h	; offset of pointer to filename portion of string 2 ?
pat2	equ	62h	; offset of pattern for filling in wildcards
			; in filename 2
str1	equ	70h	; some string (not an offset...!!?)


;	Begin.  Set disk transfer address and slash.

	org 100h	; ***

	mov	dx,dta	; offset
	mov	ah,1ah
	int	21h
	mov	ax,3700h	; set switch char
	int	21h
	cmp	dl,'/'
	jne	f1
	mov	byte [slash],'\'
f1:	cld

;	Compute buffer size.

	mov	ax,sp
	sub	ax,buf+100h	; offset
	shr	ax,1
	mov	al,0
	mov	[buflen],ax	; buffer must end 100h bytes before stack

;	Handle switches

	mov	di,81h
	mov	ch,0
	mov	cl,[di-1]

f2:	jcxz	f8		; if end of string
	mov	al,dl		; switch character
				; (NASM: 88 d0 / A86: 8a c2)
	repnz	scasb
	jne	f8		; if done searching
	cmp	al,'/'
	je	f3		; if slash, don't require preceding white space
	mov	al,[di-2]
	cmp	al,' '
	je	f3		; if space
	cmp	al,TAB
	jne	f2		; if not tab

;	Interpret switch (must be numeric argument)

f3:	xor	bx,bx		; value of the argument
				; (A86: 33 db / NASM: 31 db)
	push	cx
	mov	cx,10
	mov	si,di
f4:	mov	byte [si-1],' '
	lodsb
	sub	al,'0'
	cmp	al,9	
	ja	f5		; if not digit
	xchg	ax,bx
	mul	cx
	mov	bh,0
	add	bx,ax
	jmp	short f4	; loop back for more (short jump)

f5:	mov	[maxerrs],bx	; save argument
	pop	cx
	cmp	al,' '-'0'
	je	f2		; if ended with space
	cmp	al,TAB-'0'
	je	f2		; if ended with tab
	cmp	al,CR-'0'
	je	f2		; if ended with CR
	mov	dx,msg1		; illegal switch (offset of message)
	jmp	short f7	; error quit
f6:	mov	dx,msg2		; insufficient parameters (message offset)
f7:	jmp	errend

;	Start scanning parameters.

f8:	mov	si,81h
f9:	lodsb			; skip spaces
	cmp	al,' '
	je	f9		; if space
	cmp	al,TAB
	je	f9		; if tab
	dec	si		; get first parameter
	mov	di,str1		; offset 1st filename:
	call	getparm
	jz	f6		; if no parameter present
	pushf
	push	word [attr]	; only low byte needed
	mov	[fn1],bx	; beginning of file name 1
	lea	di,[bx+14]
	mov	[str2],di	; start of second string
	call	getparm
	mov	[fn2],bx
	mov	si,bx		; set up pat2
	mov	di,pat2		; A86 uses the offset here, too
	mov	cx,7
	rep	movsw
	pop	word [attr]	; only low byte needed
	popf
	jns	f11		; if wild cards
	cmp	word [maxerrs], byte 0
	jnz	f10		; if maxerrs given
	mov	word [maxerrs], -1	; default = -1
	mov	byte [hyph],0	; don't print hyphens
f10:	call	doit
	jmp	short quit

;	Compare multiple files.

f11:	cmp	word [maxerrs], byte 0
	jnz	f12		; if maxerrs given
	mov	word [maxerrs], 10	; default = 10
f12:	mov	dx,str1		; offset of the name
	mov	cx,1
	mov	ah,4eh		; find first file
	int	21h
	jc	f14		; if not found
	mov	word [okmsg], crlf	; re-set pointer...: not ok

f13:	mov	si,flname	; offset (*correct???*)
	mov	di,[fn1]
	mov	cx,7
	rep	movsw
	mov	dx,str1		; offset of name
	call	print
	call	doit
	mov	ah,4fh		; find next file
	int	21h
	jnc	f13

quit:	mov	ax,4c00h
	int	21h

f14:	mov	dx,msg3		; file not found
	jmp	errend

;	GETPARM	Get next parameter.  Upon entry, si points to the next character
;		in the command string and di points where to put it.  On exit,
;		si points to the next+1 character in the command line, di points
;		to the end of the parameter, and bx points to the beginning of
;		the filename part of the string.  AH=0 if no parameter, 1 if
;		wild cards are present, and 80h if it is a file.  Flags are set
;		according to AH.

getparm:
	mov	[str2],di

gp0:	lodsb		; skip separators
	cmp	al,' '
	je	getparm
	cmp	al,TAB
	je	getparm
	cmp	al,CR
	mov	ah,0
	je	gp8		; if CR

gp1:	stosb			; copy until separator
	lodsb
	cmp	al,' '		; check for separator
	je	gp2
	cmp	al,TAB
	je	gp2
	cmp	al,CR
	jne	gp1
	dec	si

;	Process the parameter.

gp2:	mov	byte [di],0
	mov	bx,di
	mov	ah,81h		; scan for start of file name

gp3:	dec	bx
	cmp	bx,[str2]
	jl	gp5		; if past beginning
	mov	al,[bx]
	cmp	al,[slash]
	je	gp6		; if '\' or '/'
	cmp	al,'?'		; check for wild cards
	je	gp4
	cmp	al,'*'
	jne	gp3
gp4:	and	ah,7fh		; clear no-wild bit
	jmp	short gp3	; (short jump)

gp5:	cmp	byte [bx+2],':'	; no dir. given; remove drive letter
	jne	gp6
	inc	bx
	inc	bx

gp6:	inc	bx
	or	ah,ah		; (A86: 0a e4 / NASM: 08 e4)
	jns	gp9		; if wild cards
	cmp	bx,di		; (A86: 3b df / NASM: 39 fb)
	mov	ah,1
	je	gp8		; if no file name
	mov	ax,4300h	; see if directory
	mov	dx,[str2]	; read pointer
	int	21h
	mov	ah,80h
	jc	gp9		; if not found
	test	cl,10h		; test attribute for directory
	jz	gp9		; if file

;	It's a directory.

	mov	al,[slash]
	stosb
	mov	ah,1

gp8:	push	ax
	mov	bx,di		; add "*.*"
	mov	ax,'*.'		; A86: '.*' (* is the low half!)
	stosw
	stosb
	mov	byte [di],0
	pop	ax

;	Return.

gp9:	or	ah,ah		; set flags (A86: 0a e4 / NASM: 08 e4)
	ret

;	DOIT	Do it.  Str1, str2, and pat2 are assumed to be set up.

doit:	mov	si,pat2		; offset of pattern
	mov	bx,[fn1]	; read pointer
	mov	di,[fn2]	; read pointer
	mov	cx,8
	call	dopart		; translate wild cards for main part of file name
	dec	si
d1:	lodsb			; skip to file extension
	or	al,al		; (A86: 0a c0 / NASM: 08 c0)
	jz	d3		; if end of file
	cmp	al,'.'
	jne	d1
	stosb			; store '.'
d2:	mov	al,byte [bx]	; skip to extension in first file name
	cmp	al,0
	jz	d2a
	inc	bx
	cmp	al,'.'
	jne	d2
d2a:	mov	cl,3		; translate wild cards for file extension
	call	dopart

;	Set up files.

d3:	mov	byte [di],0	; store terminating zero
	mov	dx,str1		; open file
	mov	ax,3d00h
	int	21h

%ifndef cloneA86
	jc	err_1		; if error (shorter than jnc... jmp)
%else
	jnc	no_err_1
	jmp	err
no_err_1:
%endif

	mov	si,ax		; (A86: 8b f0 / NASM: 89 c6)
	mov	ax,3d00h	; open file
	mov	dx,[str2]	; read pointer
	int	21h
	jnc	d3a		; if no error
	cmp	ax,2
	jz	noerr_0
	jmp	err1		; if not file-not-found
noerr_0:
	call	hyphens
	mov	dx,fil		; offset
	call	print
	mov	dx,[str2]	; read pointer
	call	print
	mov	dx,msg3+4	; ' not found.'
	call	print
	mov	ah,3eh		; close other file
	mov	bx,si
	int	21h

%ifndef cloneA86
err_1:	jc	err_2
%else
	jnc	no_err_2
	jmp	err
no_err_2:
%endif

	ret

d3a:	mov	bx,ax		; (A86: 8b d8 / NASM: 89 c3)
	mov	word [bufno], 0
	mov	word [errcnt], 0

;	Loop for comparing.  si, bx = file handles.

d4:	mov	cx,[buflen]	; read from file 1
	mov	dx,buf		; offset of big buffer
	xchg	si,bx		; (A86 / NASM swap arguments)
	mov	ah,3fh
	int	21h

%ifndef cloneA86
err_2:	jc	err_3		; if error
%else
	jnc	no_err_3
	jmp	err
no_err_3:
%endif

	mov	word [len1],ax
	push	bx
	mov	bx,si
	mov	si,dx
	add	dx,cx		; read from file 2
	mov	ah,3fh
	int	21h
err_3:	jnc	noerr_1
	jmp	err
noerr_1:
	mov	word [len2],ax
	push	bx
	mov	di,dx
	cmp	ax,[len1]
	jle	d4z		; find minimum length
	mov	ax,[len1]
d4z:	cmp	ax,[buflen]	; whether this is the last
	pushf

;	Begin loop over mini-buffers

d5:	mov	cx,256
	cmp	ax,cx		; (A86: 3b c1 / NASM: 39 c8)
	jae	d5a
	mov	cx,ax
d5a:	sub	ax,cx		; (A86: 2b c1 / NASM: 29 c8)
	push	ax

;	Do comparison.

d6:	repz	cmpsb
	je	d11		; if buffers equal

;	Print error message.

	push	cx
	push	di
	cmp	word [errcnt], byte 0
	jne	d7		; if not the first error
	mov	dx,hdr		; print header
	call	print
d7:	inc	word [errcnt]
	mov	di,lyne		; *patch in line offset here (in hex)*
	mov	dx,di
	mov	ax,[bufno]
	cmp	ah,0
	je	d8		; if not a huge file
	push	ax
	mov	al,ah
	call	hex
	stosw
	pop	ax
d8:	call	hex
	stosw
	mov	ax,si		; (A86: 8b c6 / NASM: 89 f0)
	sub	ax,buf+1
	call	hex
	stosw
	pop	bx		; old di
	mov	al,[si-1]	; convert data bytes to hex
	mov	di,lyne+6	; *patch in one hexdumped byte here*
	call	dumpbyte
	mov	al,[bx-1]
	call	dumpbyte

;	Trim spaces from input line.

d9:	dec	di
	cmp	byte [di],' '
	je	d9		; if space
	inc	di
	mov	ax,CR+256*LF
	stosw
	mov	al,0
	stosb
	mov	di,bx
	call	print		; print the line
	pop	cx
	mov	ax,[errcnt]
	cmp	ax,[maxerrs]
	jb	d10		; if within limits
	mov	dx,atleast
	call	print
	pop	ax
	popf
	pop	bx
	pop	si
	jmp	short d14

d10:	or	cx,cx		; (A86: 0b c9 / NASM: 09 c9)
	jnz	d6		; if more comparisons to do

d11:	inc	word [bufno]	; end mini-buffer
	pop	ax
	or	ax,ax		; (A86: 0b c0 / NASM: 09 c0)
	jz	nod5
	jmp	d5		; if more in this buffer
nod5:
	popf
	pop	bx		; get file handles
	pop	si
	jnz	nod4
	jmp	d4		; if not eof yet
nod4:
	mov	ax,[len2]
	sub	ax,[len1]
	mov	cl,'2'
	jg	d12		; if excess bytes on file 2
	je	d14		; if exact match
	dec	cx
	xchg	si,bx		; (NASM/A86 swap arguments)
	neg	ax

;	Excess bytes on some file.

d12:	mov	byte [xs+26],cl	; *patch in "1" or "2"*
	mov	[len1],ax
	mov	cx,[buflen]
	mov	dx,buf		; offset of big buffer

d13:	mov	ah,3fh		; read excess bytes
	int	21h
	jc	err
	add	[len1],ax
	cmp	ax,cx		; (A86: 3b c1 / NASM: 39 c8)
	je	d13		; if more to go

	mov	al,[len1+1]	; form and print excess-bytes message
	call	hex		; high byte first
	mov	[xs],ax		; *patch in hex offset, high*
	mov	al,[len1]	; now low byte
	call	hex		; *patch in hex offset, low*
	mov	[xs+2],ax
	push	bx
	mov	dx,xs
	call	print
	pop	bx
	inc	word [errcnt]
	
;	Close files and print error count.

d14:	mov	ah,3eh		; close file
	int	21h
	jc	err
	mov	ah,3eh
	xchg	si,bx		; (NASM/A86 swap arguments)
	int	21h
	jc	err

	mov	dx,[okmsg]	; get OK-message address
	mov	ax,[errcnt]
	or	ax,ax		; (A86: 0b c0 / NASM: 09 c0)
	jz	d17		; if no errors
	mov	bx,10
	xor	cx,cx		; (A86: 33 c9 / NASM: 31 c9)

d15:	xor	dx,dx		; convert to decimal (A86: 33 d2 / NASM: 31 d2)
	div	bx
	add	dl,'0'
	push	dx
	inc	cx
	or	ax,ax		; (A86: 0b c0 / NASM: 09 c0)
	jnz	d15		; if more digits
	mov	ah,2

d16:	pop	dx		; print the number
	int	21h
	loop	d16
	mov	dx,nerrs

;	Common ending routine.

d17:	jmp	short print	; call print and return

;	Process errors.

err1:	cmp	ax,5
	jne	err		; if not invalid path
	call	hyphens
	mov	dx,msg4
	jmp	short errend

err:	mov	dx,msg3		; file not found
	cmp	ax,2
	je	errend		; if file not found
	call	hyphens
	mov	dx,msg5		; I/O error
;	jmp	short errend	; (control falls through)

;	Ending stuff.

errend:	mov	bx,stderr
	call	print0
	mov	ax,4c01h
	int	21h

;	HYPHENS	Print hyphens (but only if this is a multiple comp).
;	PRINT	Print the line at (DX).  Destroys AX,CX.
;	PRINT0	Same as above, but also requires BX = output handle.

hyphens:
	mov	dx,hyph
print:	mov	bx,stdout
print0:	push	di
	mov	di,dx
	mov	cx,-1
	mov	al,0		; find length
	repnz	scasb
	not	cx
	dec	cx
	mov	ah,40h
	int	21h
	pop	di
	ret

;	DOPART	Copy string from [si] to [di], changing wild cards to those
;		present in [bx].  This copies up to CX characters.

dopart:	lodsb
	cmp	al,'.'
	je	dp5		; if end of this part
	cmp	al,0
	je	dp5		; if end
	cmp	al,'*'
	jne	dp1
	dec	si
	mov	al,'?'
dp1:	cmp	al,'?'
	jne	dp2		; if not wild card
	mov	al,[bx]
	cmp	al,'.'
	je	dp3
	cmp	al,0
	je	dp3
dp2:	stosb
dp3:	cmp	byte [bx],'.'	; advance [bx], but not past '.'
				; or end of string
	je	dp4
	cmp	byte [bx],0
	je	dp4
	inc	bx
dp4:	loop	dopart
	inc	si
dp5:	ret

;	DUMPBYTE Print a byte in hexadecimal and then as a character.

dumpbyte:
	push	ax
	mov	ax,'  '
	stosw
	pop	ax
	push	ax
	call	hex
	stosw
	mov	al,' '
	stosb
	pop	ax
	cmp	al,0
	je	db2		; if null character
	cmp	al,' '
	jb	db1		; if control character
	cmp	al,127
	ja	db2		; if not an ASCII character
	mov	ah,al
	mov	al,39		; save normal character (in single quotes)
				; char number 39 is the single quote
	stosw
	stosb
	ret

db1:	add	al,'A'-1	; control character
	mov	ah,al
	mov	al,'^'
	jmp	db3

db2:	mov	ax,'  '		; just put spaces
db3:	stosw
	mov	al,' '
	stosb
	ret

;	HEX	Convert a byte to hexadecimal.  Destroys (CL).

hex:	mov	ah,0
	mov	cl,4
	shl	ax,cl
	shr	al,cl
	add	al,90h		; these four instructions change to ascii hex
	daa
	adc	al,40h
	daa
	xchg	ah,al		; (NASM/A86 swap arguments)
	add	al,90h
	daa
	adc	al,40h
	daa
	ret

slash	db	'/'
msg1	db	'Illegal switch.',CR,LF,0
msg2	db	'Insufficient parameters.',CR,LF,0
msg3	db	'File not found.',CR,LF,0
msg4	db	'Access denied.',CR,LF,0
msg5	db	'I/O error.',CR,LF,0
fil	db	'File ',0
hyph	db	'--',0
hdr	db	CR,LF,'Byte    File 1  File 2',CR,LF,0
lyne	db	'xxxx    aa      bb "x"',CR,LF,0
xs	db	'nnnn excess bytes on file x.',CR,LF,0
nerrs	db	' error(s).',CR,LF,0
ok	db	'Files compare OK.'	; (continues...)
crlf	db	CR,LF,0
atleast	db	'At least ',0

okmsg	dw	ok
maxerrs	dw	0


dta:	; this is where "Area 43" starts

attr	equ	dta+21		; place of a byte in dta
flname	equ	dta+30		; ditto

bss	equ	dta+43
	; the next things are all words:
errcnt	equ	bss+0
bufno	equ	bss+2
len1	equ	bss+4
len2	equ	bss+6
buflen	equ	bss+8
	; finally, we have the big buffer:
buf	equ	bss+10

;	end
